@@ -32,7 +32,9 @@ enum {				/* Server data responses to be parsed;
 	DATA_RESPONSE_FETCH_FLAGS,
 	DATA_RESPONSE_FETCH_DATE,
 	DATA_RESPONSE_FETCH_SIZE,
+	DATA_RESPONSE_FETCH_STRUCTURE,
 	DATA_RESPONSE_FETCH_BODY,
+	DATA_RESPONSE_IDLE,
 };
 regexp responses[] = {		/* Server data responses to be parsed;
 				 * regular expressions patterns. */
@@ -42,7 +44,7 @@ regexp responses[] = {		/* Server data responses to be parsed;
 	{ "\\* NAMESPACE (NIL|\\(\\(\"([[:graph:]]*)\" \"([[:print:]])\"\\)"
 	  "[[:print:]]*\\)) (NIL|\\([[:print:]]*\\)) (NIL|\\([[:print:]]*\\))"
 	  "\r\n", NULL, 0, NULL },
-	{ "\\* STATUS [[:print:]]* \\(([[:alnum:] ]*)\\)\r\n", NULL, 0, NULL },
+	{ "\\* STATUS [[:print:]]* \\(([[:alnum:] ]*)\\) *\r\n", NULL, 0, NULL },
 	{ "MESSAGES ([[:digit:]]+)", NULL, 0, NULL },
 	{ "RECENT ([[:digit:]]+)", NULL, 0, NULL },
 	{ "UNSEEN ([[:digit:]]+)", NULL, 0, NULL },
@@ -56,13 +58,15 @@ regexp responses[] = {		/* Server data responses to be parsed;
 	{ "FLAGS \\(([[:print:]]*)\\)", NULL, 0, NULL },
 	{ "INTERNALDATE \"([[:print:]]*)\"", NULL, 0, NULL },
 	{ "RFC822.SIZE ([[:digit:]]+)", NULL, 0, NULL },
+	{ "BODYSTRUCTURE (\\([[:print:]]+\\))", NULL, 0, NULL },
 	{ "\\* [[:digit:]]+ FETCH \\([[:print:]]*BODY\\[[[:print:]]*\\] "
 	  "(\\{([[:digit:]]+)\\}\r\n|\"([[:print:]]*)\")", NULL, 0, NULL },
+	{ "\\* [[:digit:]]+ RECENT\r\n", NULL, 0, NULL },
 	{ NULL, NULL, 0, NULL }
 };
 
 
-int receive_response(session *ssn, char *buf);
+int receive_response(session *ssn, char *buf, long timeout, int timeoutfail);
 
 int check_tag(char *buf, session *ssn, int tag);
 int check_bye(char *buf);
@@ -74,11 +78,12 @@ int check_trycreate(char *buf);
  * Read data the server sent.
  */
 int
-receive_response(session *ssn, char *buf)
+receive_response(session *ssn, char *buf, long timeout, int timeoutfail)
 {
 	ssize_t n;
 
-	if ((n = socket_read(ssn, buf, INPUT_BUF)) == -1)
+	if ((n = socket_read(ssn, buf, INPUT_BUF, timeout ? timeout :
+	    (long)(get_option_number("timeout")), timeoutfail)) == -1)
 		return -1;
 
 
@@ -150,6 +155,21 @@ check_bye(char *buf)
 }
 
 
+/*
+ * Check if server sent a PREAUTH response (connection already authenticated
+ * by external means).
+ */
+int
+check_preauth(char *buf)
+{
+
+	if (xstrcasestr(ibuf.data, "* PREAUTH"))
+		return 1;
+	else
+		return 0;
+}
+
+
 /*
  * Check if the server sent a continuation request.
  */
@@ -194,7 +214,7 @@ response_generic(session *ssn, int tag)
 
 	do {
 		buffer_check(&ibuf, ibuf.len + INPUT_BUF);
-		if ((n = receive_response(ssn, ibuf.data + ibuf.len)) == -1)
+		if ((n = receive_response(ssn, ibuf.data + ibuf.len, 0, 1)) == -1)
 			return -1;
 		ibuf.len += n;
 
@@ -221,15 +241,15 @@ response_greeting(session *ssn)
 
 	buffer_reset(&ibuf);
 
-	if (receive_response(ssn, ibuf.data) == -1)
-		return -1;
-
-	if (check_bye(ibuf.data))
+	if (receive_response(ssn, ibuf.data, 0, 1) == -1)
 		return -1;
 
 	verbose("S (%d): %s", ssn->socket, ibuf.data);
 
-	if (xstrcasestr(ibuf.data, "* PREAUTH"))
+	if (check_bye(ibuf.data))
+		return -1;
+
+	if (check_preauth(ibuf.data))
 		return STATUS_RESPONSE_PREAUTH;
 
 	return STATUS_RESPONSE_NONE;
@@ -282,6 +302,9 @@ response_capability(session *ssn, int tag)
 		if (xstrcasestr(s, "CHILDREN"))
 			ssn->capabilities |= CAPABILITY_CHILDREN;
 
+		if (xstrcasestr(s, "IDLE"))
+			ssn->capabilities |= CAPABILITY_IDLE;
+
 		xfree(s);
 	}
 
@@ -666,9 +689,41 @@ response_fetchsize(session *ssn, int tag, char **size)
 }
 
 
+/*
+ * Process the data that server sent due to IMAP FETCH BODYSTRUCTURE client
+ * request.
+ */
+int
+response_fetchstructure(session *ssn, int tag, char **structure)
+{
+	int r;
+	char *s;
+	regexp *re;
+
+	if ((r = response_generic(ssn, tag)) == -1)
+		return -1;
+
+	re = &responses[DATA_RESPONSE_FETCH];
+	if (!regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0)) { 
+		s = xstrndup(ibuf.data + re->pmatch[1].rm_so,
+		    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
+
+		re = &responses[DATA_RESPONSE_FETCH_STRUCTURE];
+		if (!regexec(re->preg, s, re->nmatch, re->pmatch, 0)) {
+			*structure = xstrndup(s + re->pmatch[1].rm_so,
+			    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
+		}
+		xfree(s);
+	}
+
+	return r;
+}
+
+
 /*
  * Process the data that server sent due to IMAP FETCH BODY[] client request,
- * ie. FETCH BODY[HEADER], FETCH BODY[TEXT], FETCH BODY[HEADER.FIELDS ()].
+ * ie. FETCH BODY[HEADER], FETCH BODY[TEXT], FETCH BODY[HEADER.FIELDS
+ * (<fields>)], FETCH BODY[<part>].
  */
 int
 response_fetchbody(session *ssn, int tag, char **body, size_t *len)
@@ -690,7 +745,7 @@ response_fetchbody(session *ssn, int tag, char **body, size_t *len)
 
 	do {
 		buffer_check(&ibuf, ibuf.len + INPUT_BUF);
-		if ((n = receive_response(ssn, ibuf.data + ibuf.len)) == -1)
+		if ((n = receive_response(ssn, ibuf.data + ibuf.len, 0, 1)) == -1)
 			return -1;
 		ibuf.len += n;
 
@@ -725,3 +780,36 @@ response_fetchbody(session *ssn, int tag, char **body, size_t *len)
 
 	return r;
 }
+
+
+/*
+ * Process the data that server sent due to IMAP IDLE client request.
+ */
+int
+response_idle(session *ssn, int tag)
+{
+	regexp *re;
+
+	re = &responses[DATA_RESPONSE_IDLE];
+
+	do {
+		buffer_reset(&ibuf);
+
+		switch (receive_response(ssn, ibuf.data, 29 * 60, 0)) {
+		case -1:
+			return -1;
+			break; /* NOTREACHED */
+		case 0:
+			return STATUS_RESPONSE_TIMEOUT;
+			break; /* NOTREACHED */
+		}
+
+		verbose("S (%d): %s", ssn->socket, ibuf.data);
+
+		if (check_bye(ibuf.data))
+			return -1;
+
+	} while (regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0));
+
+	return STATUS_RESPONSE_UNTAGGED;
+}
