@@ -123,7 +123,7 @@ open_secure_connection(session *ssn, const char *server, const char *port,
 		    server, port, ERR_error_string(ERR_get_error(), NULL));
 		goto fail;
 	}
-	if (get_cert(ssn) == -1)
+	if (get_option_boolean("certificates") && get_cert(ssn) == -1)
 		goto fail;
 
 	SSL_CTX_free(ctx);
@@ -187,9 +187,9 @@ close_secure_connection(session *ssn)
  * Read data from socket.
  */
 ssize_t
-socket_read(session *ssn, char *buf, size_t len)
+socket_read(session *ssn, char *buf, size_t len, long timeout, int timeoutfail)
 {
-	int s, t;
+	int s;
 	ssize_t r;
 	fd_set fds;
 
@@ -203,9 +203,8 @@ socket_read(session *ssn, char *buf, size_t len)
 
 	memset(buf, 0, len + 1);
 
-	t = (int)(get_option_number("timeout"));
-	if (t > 0) {
-		tv.tv_sec = t;
+	if (timeout > 0) {
+		tv.tv_sec = timeout;
 		tv.tv_usec = 0;
 		tvp = &tv;
 	}
@@ -242,7 +241,7 @@ socket_read(session *ssn, char *buf, size_t len)
 	if (s == -1) {
 		error("waiting to read from socket; %s\n", strerror(errno));
 		goto fail;
-	} else if (s == 0) {
+	} else if (s == 0 && timeoutfail) {
 		error("timeout period expired while waiting to read data\n");
 		goto fail;
 	}
@@ -298,31 +297,18 @@ socket_secure_read(session *ssn, char *buf, size_t len)
 ssize_t
 socket_write(session *ssn, const char *buf, size_t len)
 {
-	int s, t;
+	int s;
 	ssize_t w, wt;
 	fd_set fds;
 
-	struct timeval tv;
-
-	struct timeval *tvp = NULL;
-
 	w = wt = 0;
 	s = 1;
 
-	t = (int)(get_option_number("timeout"));
-	if (t > 0)
-		tvp = &tv;
-
 	FD_ZERO(&fds);
 	FD_SET(ssn->socket, &fds);
 
 	while (len) {
-		if (t > 0) {
-			tv.tv_sec = t;
-			tv.tv_usec = 0;
-		}
-
-		if ((s = select(ssn->socket + 1, NULL, &fds, NULL, tvp) > 0 &&
+		if ((s = select(ssn->socket + 1, NULL, &fds, NULL, NULL) > 0 &&
 		    FD_ISSET(ssn->socket, &fds))) {
 #ifndef NO_SSLTLS
 			if (ssn->ssl) {
